﻿using System;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;

namespace ThreadingBenchmark
{
    internal class Program
    {
        private const int STEPS = 10;
        private const int STEPSIZE = 1000000;
        private const int REPETITIONSPERSTEP = 3;
        private const int THREADS = 4;

        private static int numberOfCalculations = 0;

        static void Main(string[] args)
        {
            Console.WriteLine("Steps: {0}", STEPS);
            Console.WriteLine("Stepsize: {0}", STEPSIZE);
            Console.WriteLine("Repetitions per step: {0}", REPETITIONSPERSTEP);
            Console.WriteLine("Threads: {0}", THREADS);

            Console.WriteLine(string.Empty);
            Console.WriteLine("CheapCalculation");
            Console.WriteLine(string.Empty);

            Console.Write("\t");
            for (int i = 0; i <= STEPS; i++)
            {
                Console.Write(i * STEPSIZE + ",\t");
            }

            ExecuteTask("Single Thread", UseSingleThread, CheapCalculation);
            ExecuteTask("Multiple Threads", UseMultipleThreads, CheapCalculation);
            ExecuteTask("ThreadPool", UseThreadPool, CheapCalculation);
            ExecuteTask("ParallelAPI (TPL)", UseParallelAPI, CheapCalculation);

            Console.WriteLine(string.Empty);
            Console.WriteLine(string.Empty);
            Console.WriteLine("ExpensiveCalculation");

            ExecuteTask("Single Thread", UseSingleThread, ExpensiveCalculation);
            ExecuteTask("Multiple Threads", UseMultipleThreads, ExpensiveCalculation);
            ExecuteTask("ThreadPool", UseThreadPool, ExpensiveCalculation);
            ExecuteTask("ParallelAPI (TPL)", UseParallelAPI, ExpensiveCalculation);
            Console.ReadKey();
        }

        private static void ExecuteTask(string name, Action<Action<int>> task, Action<int> calculation)
        {
            Console.WriteLine(string.Empty);
            Console.Write("{0}\t", name);
            for (int i = 0; i <= STEPS; i++)
            {
                numberOfCalculations = i * STEPSIZE;
                Console.Write(GetExecutionTime(task, calculation) + ",\t");
            }            
        }

        private static long GetExecutionTime(Action<Action<int>> task, Action<int> calculation)
        {
            var stopwatch = new System.Diagnostics.Stopwatch();
            stopwatch.Start();

            for (int i = 0; i < REPETITIONSPERSTEP; i++)
            {
                task(calculation);
            }            

            stopwatch.Stop();

            return stopwatch.ElapsedMilliseconds / REPETITIONSPERSTEP;
        }

        private static void UseSingleThread(Action<int> action)
        {
            for (int i = 0; i < numberOfCalculations; i++)
            {
                action(i);
            }
        }

        private static void UseMultipleThreads(Action<int> action)
        {
            object lockObject = new object();
            var workerThreads = new List<Thread>();
            int currentNumber = 1;

            for (int i = 0; i < THREADS; i++)
            {
                var thread = new Thread(delegate()
                {
                    int number = 0;

                    while (true)
                    {
                        lock (lockObject)
                        {
                            if (currentNumber < numberOfCalculations)
                            {
                                number = ++currentNumber;
                            }
                            else
                            {
                                break;
                            }
                        }

                        action(number);
                    }
                });

                workerThreads.Add(thread);
            }

            foreach (var thread in workerThreads)
            {
                thread.Start();
            }

            foreach (var thread in workerThreads)
            {
                thread.Join();
            }
        }

        private static void UseThreadPool(Action<int> action)
        {
            for (int i = 0; i < numberOfCalculations; i++)
            {
                ThreadPool.QueueUserWorkItem(delegate { action(i); });
            }
        }

        private static void UseParallelAPI(Action<int> action)
        {
            Parallel.For(0, numberOfCalculations, n => action(n));
        }

        private static void CheapCalculation(int number)
        {            
        }

        private static void ExpensiveCalculation(int number)
        {
            Thread.SpinWait(50);
        }
    }
}
